/**

	Mindev Engine
	v1.0.0

	Created by Nathan Starkey
	Compiled Tue 14/04/2020

**/



class Animations {
    constructor(){
        this.animations = [];
    }


    update(delta){
        for(var i=0; i<this.animations.length; i++){
            var animation = this.animations[i];
            if(!animation.running) continue;

            animation.time += animation.velocity * delta;

            var progress = Math.max(0, Math.min(1, animation.time/animation.length));
			animation.progress = progress;

            if(animation.update){
                animation.update(progress);
            }

            if(progress===1 && animation.repeat==="once"){
                this.remove(animation);
            }
			
            if((progress===0 || progress===1) && animation.repeat==="rebound"){
                animation.velocity *= -1;
            }

            if(progress===1 && animation.repeat==="infinite"){
                animation.time = 0;
            }
        }
    }


    draw(){
        for(var i=0; i<this.animations.length; i++){
            var animation = this.animations[i];

            if(animation.draw){
                animation.draw(animation.progress);
            }
        }
    }


    add(update, draw, length, repeat){
        var animation = {
            running: true,
            repeat: repeat || "once",

            update: update,
            draw: draw,

            time: 0,
            length: length,
            progress: 0,
            velocity: 1
        };

        this.animations.push(animation);

        return animation;
    }


    remove(animation){
		if(!animation.running) return;
        animation.running = false;
        this.animations.splice(this.animations.indexOf(animation), 1);
    }


    play(animation){
        animation.running = true;
    }


    pause(animation){
        animation.running = false;
    }


    restart(animation){
        animation.time = 0;
    }
}



class BoxPhysics2D {
    constructor(){

    }


    testCollision(box1, box2){
        var b1v = [
            this.rotatePoint(box1.x, box1.y, box1.r, box1.ox, box1.oy),
            this.rotatePoint(box1.x + box1.width, box1.y, box1.r, box1.ox, box1.oy),
            this.rotatePoint(box1.x + box1.width, box1.y + box1.height, box1.r, box1.ox, box1.oy),
            this.rotatePoint(box1.x, box1.y + box1.height, box1.r, box1.ox, box1.oy)
        ], b2v = [
            this.rotatePoint(box2.x, box2.y, box2.r, box2.ox, box2.oy),
            this.rotatePoint(box2.x + box2.width, box2.y, box2.r, box2.ox, box2.oy),
            this.rotatePoint(box2.x + box2.width, box2.y + box2.height, box2.r, box2.ox, box2.oy),
            this.rotatePoint(box2.x, box2.y + box2.height, box2.r, box2.ox, box2.oy)
        ];


        if(window.game){
            var mw = 1,
                mh = 1;

            game.context.fillStyle = "#f00";
            game.context.fillRect(b1v[0][0] - mw/2, b1v[0][1] - mh/2, mw, mh);
            game.context.fillRect(b1v[1][0] - mw/2, b1v[1][1] - mh/2, mw, mh);
            game.context.fillRect(b1v[2][0] - mw/2, b1v[2][1] - mh/2, mw, mh);
            game.context.fillRect(b1v[3][0] - mw/2, b1v[3][1] - mh/2, mw, mh);

            game.context.fillStyle = "#f80";
            game.context.fillRect(b2v[0][0] - mw/2, b2v[0][1] - mh/2, mw, mh);
            game.context.fillRect(b2v[1][0] - mw/2, b2v[1][1] - mh/2, mw, mh);
            game.context.fillRect(b2v[2][0] - mw/2, b2v[2][1] - mh/2, mw, mh);
            game.context.fillRect(b2v[3][0] - mw/2, b2v[3][1] - mh/2, mw, mh);
        }


        var axes = [
            ...this.getAxes(b1v),
            ...this.getAxes(b2v)
        ];


        var overlap,
            axis;

        for(var i=0, n=axes.length; i<n; i++){
            var b1Min, b1Max,
                b2Min, b2Max;

            for(var j=0, o=b1v.length; j<o; j++){
                var k = b1v[j][0] * axes[i][0] + b1v[j][1] * axes[i][1];

                if(j===0 || b1Min > k) b1Min = k;
                if(j===0 || b1Max < k) b1Max = k;
            }

            for(var j=0, o=b2v.length; j<o; j++){
                var k = b2v[j][0] * axes[i][0] + b2v[j][1] * axes[i][1];

                if(j===0 || b2Min > k) b2Min = k;
                if(j===0 || b2Max < k) b2Max = k;
            }

            if(b1Max < b2Min || b1Min > b2Max) return false;

            else {
                var o = b1Max > b2Min ? b1Max - b2Min : b2Max - b1Min;

                if(overlap===undefined || o < overlap){
                    overlap = o;
                    axis = axes[i]
                }
            }
        }

        return {
            overlap,
            axis
        };
    }


    rotatePoint(x, y, r, ox, oy){
        var theta = r / 180 * Math.PI;

        x -= ox;
        y -= oy;

        var tx = x * Math.cos(theta) - y * Math.sin(theta),
            ty = x * Math.sin(theta) + y * Math.cos(theta);

        tx += ox;
        ty += oy;

        return [ tx, ty ];
    }


    getAxes(v){
        var v1x = v[0][0] - v[1][0],
            v1y = v[0][1] - v[1][1],
            v2x = v[1][0] - v[2][0],
            v2y = v[1][1] - v[2][1],

            v1 = [ -v1y, v1x ],
            v2 = [ -v2y, v2x ],

            n1 = Math.hypot(v1[0], v1[1]),
            n2 = Math.hypot(v2[0], v2[1]);

        v1[0] /= n1;
        v1[1] /= n1;
        v2[0] /= n2;
        v2[1] /= n2;

        return [ v1, v2 ];
    }


    project(axis, shape){
        return projection;
    }
}



class FixedRatio {
    constructor(width, height, onChange){
        this.width = width;
        this.height = height;
        this.onChange = onChange;

        var that = this;

        window.addEventListener("resize", function(){
            that.onResize();
        });

        this.onResize();
    }

    setRatio(width, height){
        this.width = height;
        this.height = height;
    }

    onResize(){
        this.scale = Math.min(innerWidth / this.width, innerHeight / this.height);
        this.screenLeft = (innerWidth  - this.width * this.scale) / 2;
        this.screenTop = (innerHeight - this.height * this.scale) / 2;
        this.screenWidth = this.width * this.scale;
        this.screenHeight = this.height * this.scale;

        this.onChange ? this.onChange(this.screenLeft, this.screenTop, this.screenWidth, this.screenHeight) : 0;
    }
}



class GenericLoop {
    constructor(delta, update, draw, onFpsUpdate){
        this.delta = delta;

        this.running = false;
        this.frame = 0;
        this.last = 0;
        this.lag = 0;

        this.fpsLastUpdate = 0;
        this.fpsInterval = 2;
        this.fpsFrame = 0;
        this.fps = 0;

        if(update) this.update = update;
        if(draw) this.draw = draw;
        if(onFpsUpdate) this.onFpsUpdate = onFpsUpdate;
    }


    update(delta){}
    draw(){}
    onFpsUpdate(fps){}


    start(){
		this.fpsLastUpdate = performance.now() / 1000;
        this.running = true;
        this.loop();
    }


    stop(){
        this.running = false;
    }


    loop(){
        if(!this.running) return;

        var that = this,

            now = performance.now() / 1000,
            elapsed = now - this.last;

        this.lag += elapsed;
        this.last = now;
		
        if(now - this.fpsLastUpdate >= this.fpsInterval){
            this.fps = Math.floor(this.fpsFrame / this.fpsInterval);
            this.onFpsUpdate(this.fps);
            this.fpsLastUpdate = now;
            this.fpsFrame = 0;
        }

        while(this.lag>=this.delta){
            this.update(this.delta);
            this.lag -= this.delta;
        }

        this.fpsFrame++;
        this.frame++;
        this.draw(this.lag);

        if(this.running) requestAnimationFrame(function(){ that.loop(); });
    }
}



class InputHandler {
    constructor(ratio){
        this.keys = {};

        this.mouse = {
            x: 0,
            y: 0,
            down: false,
			touchLeft: { x: 0, y: 0, down: false },
			touchRight: { x: 0, y: 0, down: false },
			touchLast: { x: 0, y: 0, down: false }
        };

        var that = this;

        window.addEventListener("keydown", function(e){ if(!e.repeat) that.keys[e.which] = true; });
        window.addEventListener("keyup", function(e){ delete that.keys[e.which]; });

        window.addEventListener("mousemove", function(e){
            ratio ?
            (   that.mouse.x = (e.pageX - ratio.screenLeft) / ratio.screenWidth * ratio.width,
                that.mouse.y = (e.pageY - ratio.screenTop) / ratio.screenHeight * ratio.height) :
            (   that.mouse.x = e.pageX,
                that.mouse.y = e.pageY);
        });

        window.addEventListener("mousedown", function(e){ that.mouse.down = true; });
        window.addEventListener("mouseup", function(e){ that.mouse.down = false; });
		
		/*window.addEventListener("touchstart", touchMove);
		window.addEventListener("touchmove", touchMove);
		window.addEventListener("touchend", touchMove);*/
		/*
		function touchMove(e){
			that.mouse.touchLeft.down = false;
			that.mouse.touchRight.down = false;
			that.mouse.touchLast.down = false;
			
			for(var i=0; i<e.touches.length; i++){
				var pageX = e.touches[i].pageX,
				pageY = e.touches[i].pageY;
				
				if(0.5 > (pageX - ratio.screenLeft) / ratio.screenWidth / 2){
					that.touchLeft.x = (pageX - ratio.screenLeft) / ratio.screenWidth * ratio.width;
					that.touchLeft.y = (pageY - ratio.screenTop) / ratio.screenHeight * ratio.height;
				}
				
				else{
					that.touchRight.x = (pageX - ratio.screenLeft) / ratio.screenWidth * ratio.width;
					that.touchRight.y = (pageY - ratio.screenTop) / ratio.screenHeight * ratio.height;
				}
				
				if(i === e.touches.length - 1){
					that.mouse.x = (pageX - ratio.screenLeft) / ratio.screenWidth * ratio.width;
					that.mouse.y = (pageY - ratio.screenTop) / ratio.screenHeight * ratio.height;
				}
			}
		}*/

        window.addEventListener("blur", function(){
            for(var i in that.keys) delete that.keys[i];
            that.mouse.down = false;
        });
    }
}



class Scene {
    constructor(){
        this.objects = [];
        this.particles = [];

        this.player = null;
        this.players = [];
        this.rules = {};
    }


    addObject(object){ return this.objects.push(object); }
    hasObject(object){ return this.objects.includes(object); }
    removeObject(object){ this.objects.splice(this.objects.indexOf(object), 1); }
    removeAllObjects(){ this.objects.splice(0, this.objects.length); }


    addParticle(particle){ return this.particles.push(particle); }
    hasParticle(particle){ return this.particles.includes(particle); }
    removeParticle(object){ this.particles.splice(this.particles.indexOf(particle), 1); }
    removeAllParticles(){ this.particles.splice(0, this.particles.length); }

    setPlayer(player){ return this.player = player; }
    addPlayer(player){ return this.players.push(player); }
    hasPlayer(player){ return this.player.includes(player); }
    removePlayer(player){ player ? this.players.splice(this.players.indexOf(player), 1) : (this.player = null); }

    setRule(rule, value){ return this.rules[rule] = value; }
    setRules(rules){ for(var rule in rules) this.rules[rule] = rules[i]; }
    getRule(rule){ return this.rules[rule]; }
    removeRule(rule){ delete this.rules[rule]; }
    removeAllRules(){ for(var rule in this.rules) delete this.rules[rule]; }


    updateObjects(){
        for(var i=0; i<this.objects.length; i++){
            if(this.objects[i].update(...arguments)<=0){
                this.objects.splice(i, 1);
                i--;
            }
        }
    }


    updateParticles(){
        for(var i=0; i<this.particles.length; i++){
            if(this.particles[i].update(...arguments)<=0){
                this.particles.splice(i, 1);
                i--;
            }
        }
    }


    drawObjects(){
        for(var i=0; i<this.objects.length; i++){
            this.objects[i].draw(...arguments);
        }
    }


    drawParticlesBg(){
        for(var i=0; i<this.particles.length; i++){
            if(this.particles[i].layer==="background") this.particles[i].draw(...arguments);
        }
    }


    drawParticlesFg(){
        for(var i=0; i<this.particles.length; i++){
            if(this.particles[i].layer==="foreground") this.particles[i].draw(...arguments);
        }
    }
}



function Texture(src){
    var image = new Image();
    image.src = src;
    
    return image;
}



function CanvasTexture(texture, arg2){
	var canvas = document.createElement("canvas"),
		context = canvas.getContext("2d");
	
	canvas.width = 1;
	canvas.height = 1;
	canvas.context = context;
	
	if(texture.tagName === "IMG"){
		canvas.width = texture.width;
		canvas.height = texture.height;
		context.drawImage(texture, 0, 0);
	}
	
	else if(typeof texture === "number"){
		canvas.width = texture;
		canvas.height = arg2;
	}
	
	else {
		canvas.width = texture.width;
		canvas.height = texture.height;
		context.drawImage(texture, 0, 0);
	}
	
	return canvas;
}



/**
	The following bs is not game-specific, and unrelated to the Mindev engine.
	I just couldn't be bothered putting it in a seperate script.
**/

function doCollision(a){
	if(!a.isAlive) return;
	
	var game = a.game;
	
	var collisionX = a.x,
		collisionY = a.y;
	
	collisionX += a.collision.offsetX *  Math.sin((a.r - 90)/180*Math.PI);
	collisionY += a.collision.offsetX * -Math.cos((a.r - 90)/180*Math.PI);
	
	collisionX += a.collision.offsetY *  Math.sin(a.r/180*Math.PI);
	collisionY += a.collision.offsetY * -Math.cos(a.r/180*Math.PI);
	
	for(var i=0; i<game.scene.objects.length; i++){
		var b = game.scene.objects[i];
		
		if(a.parent !== b && a.collision.with.includes(b.constructor)){
			// Test to see if the objects are 'active'
			var distance = Math.hypot(collisionX - b.x, collisionY - b.y);
			if(distance >= a.collision.radius + b.collision.radius) continue;
			
			for(var areaY=Math.floor(-a.collision.height/2); areaY<=a.collision.height/2; areaY++){
				for(var areaX=Math.floor(-a.collision.width/2); areaX<=a.collision.width/2; areaX++){
					// Test to see if they actually collide
					var doesCollide = game.physics.testCollision({
						x: a.x - a.width/2,
						y: a.y - a.height/2,
						r: a.r, ox: a.x, oy: a.y,
						width : a.width,
						height: a.height
					}, {
						x: b.x - b.width/2,
						y: b.y - b.height/2,
						r: b.r, ox: b.x, oy: b.y,
						width : b.width,
						height: b.height
					});
					
					if(!doesCollide) continue;
					
					var result = detatchAt(a, b, collisionX + areaX, collisionY + areaY);
					
					if(result===1) continue;
					if(result===0) return;
				}
			}
		}
	}
}



function detatchAt(a, b, collisionX, collisionY){
	// Detatch the pixel that we've hit
	var game = a.game,
		pos = game.physics.rotatePoint(collisionX, collisionY, -b.r, b.x, b.y),
		
		pixelX = Math.floor((pos[0] - b.x + b.width /2) / b.scale),
		pixelY = Math.floor((pos[1] - b.y + b.height/2) / b.scale);
	
	if(a.collision.reaction) a.collision.reaction.call(a, collisionX, collisionY, b, pixelX, pixelY);
	
	if(((!a.collision.piercing && a.life===a.maxLife) || a.collision.isDud || (false && b.isBoss && Math.random() > 0.05)) && b.structure.data[pixelY] && b.structure.data[pixelY][pixelX]){
		a.isAlive = false;
		return a.collision.piercing ? 2 : 0;
	}
	
	
	var doesHit = b.detatch(pixelX, pixelY),
		
		speed = 0.05 / 9,
		vx = speed * Math.sin(a.r/180*Math.PI),
		vy = speed * -Math.cos(a.r/180*Math.PI);
	
	if(!doesHit) return 1;
	
	a.isAlive = false;
	
	var hitRadius = a.collision.hitRadius;
	
	for(var y = pixelY - hitRadius; y < pixelY + hitRadius; y++){
		for(var x = pixelX - hitRadius; x < pixelX + hitRadius; x++){
			if(Math.hypot(y - pixelY, x - pixelX) > hitRadius) continue;
			
			// Detatch each pixel hit
			b.detatch(Math.floor(x), Math.floor(y), vx, vy);
		}
	}
	
	if(!a.collision.piercing) return 0;
	
	return 2;
}



/*class LongTask {
	constructor(max){
		this.pending = [];
		this.index = 0;
		this.max = max;
	}
	
	
	run(){
		this.index = 0;
		
		while(this.index < this.max && this.pending[0]){
			var task = this.pending[0].splice(0, 1)[0];
			task[0].call(window, ...task[1]);
		}
	}
	
	
	call(fn, ...args){
		if(this.index >= this.max){
			this.pending.push([ fn, ...args ]);
		}
		
		else{
			fn.call(window, ...args);
			this.index++;
		}
	}
}*/



function ease(t){ return t<.5?2*t*t:-1+(4-2*t)*t; }
function mod(i, max){ return (Math.ceil(-i/max) * max + i) % max; }
function rebound(val, max){ return Math.floor(val/max) % 2 ? max - mod(val, max) : mod(val , max); }



function strongBurnReaction(x, y, target, pixelX, pixelY){
	if(this.life===this.maxLife) return;
	
	var r = Math.floor(Math.random() * 100),
		g = Math.floor(Math.random() * r),
		b = 0;
	
	burn(target, pixelX, pixelY, 3, r, g, b, 0.2);
	
	
	var r = Math.floor(Math.random() * 128 + 128),
		g = r/2,
		b = 0;
	
	burn(target, pixelX, pixelY, 2, r, g, b, 1);
}



function burnReaction(x, y, target, pixelX, pixelY){
	if(this.life===this.maxLife) return;
	
	var r = Math.floor(Math.random() * 100),
		g = Math.floor(Math.random() * r),
		b = 0;
	
	burn(target, pixelX, pixelY, 3, r, g, b, 0.2);
	
	
	var r = Math.floor(Math.random() * 128 + 128),
		g = r/2,
		b = 0;
	
	burn(target, pixelX, pixelY, 1, r, g, b, 1);
}


function burn(target, pixelX, pixelY, hitRadius, r, g, b, a){
	if(target.invincible) return;
	
	target.texture.context.fillStyle = "rgb(" + r + "," + g + "," + b + ")";
	target.texture.context.globalAlpha = a;
	
	for(var y = pixelY - hitRadius; y < pixelY + hitRadius; y++){
		for(var x = pixelX - hitRadius; x < pixelX + hitRadius; x++){
			if(
				Math.hypot(y - pixelY, x - pixelX) <= hitRadius &&
				target.structure.data[y] &&
				target.structure.data[y][x]
			){
				if(
					target.structure.colors[target.structure.data[y][x]] !== "156,242,150,1" &&
					target.structure.colors[target.structure.data[y][x]] !== "195,243,192,1" &&
					target.structure.colors[target.structure.data[y][x]] !== "156,242,151,1" &&
					target.structure.colors[target.structure.data[y][x]] !== "195,243,193,1"
				){
					target.texture.context.fillRect(x, y, 1, 1);
				}
			}
		}
	}
}



function drawText(context, textures, atlas, font, text, sx, sy, scale, wrap){
	var x = sx,
		y = sy;
	
	scale = scale || 1;
	
	for(var i=0; i<text.length; i++){
		var map = font[atlas.gui.fontMap.indexOf(text[i])];
		
		if(map){
			context.drawImage(textures.fonts, map[0], map[1], map[2], map[3], x, y, map[2] * scale, map[3] * scale);
			x += (map[2] + 1) * scale;
		}
		
		if((wrap && x > wrap) || text[i] === "\n"){
			x = sx;
			y += (font[0][3] + 1) * scale;
		}
	}
}



function touching(a, b){ return a.x<b.x+b.width&&a.x+a.width>b.x&&a.y<b.y+b.height&&a.height+a.y>b.y; }



function touchingText(that, element){
	return touching({
		x: element.offsetX,
		y: element.offsetY,
		width: element.size.width,
		height: element.size.height
	}, {
		x: that.input.mouse.x / that.scale,
		y: that.input.mouse.y / that.scale,
		width: 0,
		height: 0
	});
}



function textSize(game, text, font, scale){
	var width = 0,
		cw = 0;
		height = font[0][3];
	
	for(var i=0; i<text.length; i++){
		var map = font[game.atlas.gui.fontMap.indexOf(text[i])] || font[font.length-1];
		if(text[i] === "\n") height += map[3] + 1, cw = 0;
		else{
			cw += map[2] + 1;
			width = Math.max(width, cw);
		}
	}
	
	return { width: width * scale, height: height * scale };
}
